/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods_clock.c - Model-specific routines dealing with programmable
**                   clock generators.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

static MethodResult
methodClockWordIcd2061(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      fref,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  static const unsigned long mclkRange[16][2] = {
    { 49990000,  51000000 },
    { 50990000,  53200000 },
    { 53190000,  58500000 },
    { 58490000,  60700000 },
    { 60690000,  64400000 },
    { 60390000,  66800000 },
    { 66790000,  73500000 },
    { 73490000,  75600000 },
    { 75590000,  80900000 },
    { 80890000,  83200000 },
    { 83190000,  91500000 },
    { 91490000, 100000000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 }
  };
  os_uint32  W;          /* programming word for ICD2061 */
  os_uint32  fvco = 0;
  os_uint32  PQERR;      /* Error so far */
  os_uint32  temp;
  os_uint32  act = 0;
  os_uint32  I;          /* VCO range */
  os_uint32  D;          /* Post-VCO divider */
  os_uint32  P;          /* VCO mult */
  os_uint32  BP = 0;     /* Best 'P' so far */
  os_uint32  Q;          /* VCO divide */
  os_uint32  BQ = 0;     /* Best 'Q' so far */
  os_uint32  QMin, QMax; /* Min/max allowable Q */
  os_uint32  PMax;       /* Max allowable P */
  os_boolean found = FALSE;

  if (freq < 400000 || freq > 100000000 || clockIndex >= 2) {
    return MethodResult_InvalidParameter;
  }

  /* First find a post-VCO divisor */
  for (D = 0; D < 8; D++) {
    fvco = freq * (1 << D);
    if (fvco > 49990000) {
      break;
    }
  }

  for (I = 0; I < 16; I++) {
    if ((fvco > mclkRange[I][0]) && (fvco < mclkRange[I][1])) {
      break;
    }
  }

  /* Brute force P and Q guessing */
  PQERR = 1000000;
  QMax = (fref / 200000) - 1;
  QMin = (fref / 5000000) + 1;
  PMax = admxrcDriverParams.icd2061PMax;
  for (Q = QMin; Q <= QMax; Q++) {
    for (P = 4; P < PMax; P++) {
      os_uint32 err;

      /*
      ** The associativity of multiplication and divide operators is
      ** left to right. IF THIS FORMULA IS MODIFIED, ENSURE THAT IT
      ** CANNOT OVERFLOW AN (os_uint32).
      */
      temp = fref * P / Q / (1 << D) * 2;
      if (temp > freq) {
	err = temp - freq;
      } else {
	err = freq - temp;
      }
      if (err < PQERR) {
	BQ = Q;
	BP = P;
	PQERR = err;
	act = temp;
	found = TRUE;
      }
    }
  }

  if (!found) {
    return MethodResult_GeneralFailure;
  }

  /* Assemble the completed word apart from the register selection */
  W = (I << 17) | ((BP - 3) << 10) | (D << 7) | (BQ - 2);
	
  /*
  ** Add the ICD2061 register index
  **
  ** index = 0 means MCLK register
  ** index = 1 means VCLK1 register
  */
  switch (clockIndex) {
  case 0:
    W |= (3 << 21);  /* ICD2061's MCLK */
    break;
  case 1:	
    W |= (0 << 21);  /* ICD2061's VCLK1 */
    break;
  default:
    break;
  }

  *actual = act;
  *clockWord = W;

  return MethodResult_Success;
}

/*
** RDW' = RDW + 2;
** VDW' = VDW + 8;
**
** When changing any arithmetic expressions in this function, be careful
** not to allow arithmetic overflow of an unsigned 32-bit quantity.
*/
static MethodResult
methodClockWordIcs307(
  DeviceContext* pDevCtx,
  os_boolean     bDouble,
  os_uint32      fref,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  const os_uint32 TTL = 0x1U;
  const os_uint32 F = 0x3U;
  const os_uint32 C = 0x0U;
  os_uint32       W; /* programming word for ICS307 */
  os_uint32       max_RDWd;
  os_uint32       min_VDWd;
  os_uint32       max_VDWd;
  os_uint32       ODd;
  os_uint32       RDWd;
  os_uint32       VDWd;
  os_uint32       best = 0;
  os_uint32       best_error;
  os_uint32       best_RDW = 0;
  os_uint32       best_VDW = 0;
  os_uint32       best_ODd = 0;
  os_uint32       best_OD = 0;
  
  if (bDouble) {
    /*
    ** We're using CLK2 as the local bus clock, which is
    ** CLK1/2. So we have to multiply the desired local bus
    ** clock by 2 before programming the CLK1 frequency into
    ** the ICS307
    */
    if (freq < 5999000 || freq > 100001000) {
      return MethodResult_InvalidParameter;
    }
    freq = freq << 1;
  } else {
    if (freq < 5999000 || freq > 200001000) {
      return MethodResult_InvalidParameter;
    }
  }

  /*
  ** Get the reference divider word.
  */
  max_RDWd = fref / 200000;
  if (max_RDWd > 129) {
    max_RDWd = 129;
  }
  
  /*
  ** Brute force guessing of best OD, RDW and VDW combination!
  */
  best_error = 0xFFFFFFFFU;
  for (ODd = 2; ODd <= 10; ODd++) {
    if (ODd == 9) {
      continue;
    }
    
    for (RDWd = 3; RDWd <= max_RDWd; RDWd++) {
      max_VDWd = 20000000 * RDWd / (fref / 10);
      if (max_VDWd > 519) {
	max_VDWd = 519;
      }
      min_VDWd = 27500000 * RDWd / fref;
      if (min_VDWd < 12) {
	min_VDWd = 12;
      }
      
      for (VDWd = min_VDWd; VDWd <= max_VDWd; VDWd++) {
	os_uint32 f, error;
	
	f = fref * 2 / RDWd * VDWd / ODd;
	error = (f > freq) ? (f - freq) : (freq - f);
	if (error < best_error) {
	  best = f;
	  best_error = error;
	  best_RDW = RDWd - 2;
	  best_VDW = VDWd - 8;
	  best_ODd = ODd;
	}
      }
    }
  }
  
  W = 0;
  W |= (best_RDW & 0x7fU)  << 0;
  W |= (best_VDW & 0x1ffU) << 7;
  switch (best_ODd) {
  case 10:
    best_OD = 0;
    break;

  case 2:
    best_OD = 1;
    break;

  case 8:
    best_OD = 2;
    break;

  case 4:
    best_OD = 3;
    break;

  case 5:
    best_OD = 4;
    break;

  case 7:
    best_OD = 5;
    break;

  case 3:
    best_OD = 6;
    break;

  case 6:
    best_OD = 7;
    break;
  }
  W |= (best_OD & 0x1ffU) << 16;
  W |= (F & 0x3U) << 19;
  W |= (TTL & 0x1U) << 21;
  W |= (C & 0x3U) << 22;
  
  if (bDouble) {
    *actual = best >> 1;
  } else {
    *actual = best;
  }
  *clockWord = W;
  
  return MethodResult_Success;
}

static MethodResult
methodClockWordIcs8430_61(
  DeviceContext* pDevCtx,
  os_uint32 refFreq,
  os_uint32 reqFreq,
  os_uint32* actualFreq,
  os_uint32* clockWord)
{
  const unsigned long nval[8] = {
    2, 3, 4, 6, 8, 12, 16, 24
  };
  const os_uint32 nrange[8][2] = {
    { 250000000, 500001000 }, /* min, max freq */
    { 166660000, 333330000 },
    { 125000000, 250000000 },
    {  83330000, 166660000 },
    {  62500000, 125000000 },
    {  41660000,  83330000 },
    {  31250000,  62500000 },
    {         0,  41660000 }
  };
  const unsigned long T = 0x2U;
  unsigned long m, n;
  unsigned long bestM = 0, bestN = 0;
  os_uint32 f, bestFreq = 0; /* prevent compiler whining */
  os_uint32 err, bestErr = 0x7fffffffU;
  
  if (reqFreq < 32550000 || reqFreq > 500001000) {
    return MethodResult_InvalidParameter;
  }
  
  for (m = 250; m <= 500; m++) {
    for (n = 0; n < 8; n++) {
      /*
      ** Do not modify the order of operations in this formula unless
      ** you ensure that the new order cannot result in arithmetic
      ** overflow.
      */
      f = refFreq / (8 * nval[n]) * m;
      if (f < nrange[n][0] || f > nrange[n][1]) {
	/* Output frequency is outside limits for N current value */
	continue;
      }
      if (f > reqFreq) {
	err = f - reqFreq;
      } else {
	err = reqFreq - f;
      }
      if (err < bestErr) {
	bestErr = err;
	bestFreq = f;
	bestM = m;
	bestN = n;
      }
    }
  }
  
  if (bestM == 0) {
    /* Failed to find a suitable value */
    return MethodResult_InvalidParameter;
  }
  
  *actualFreq = bestFreq;
  *clockWord = (T << 12) | (bestN << 9) | (bestM << 0);
  
  return MethodResult_Success;
}

/*
** Bits 0-16 of the returned clockWord are the serial clock programming word
** for the ICS843034_01, but bit 17 is the reference clock select, which is
** driven to the SEL pins of the ICS843034_01 by ADB1-XRC4FX.
*/
static MethodResult
methodClockWordIcs843034_01(
  DeviceContext* pDevCtx,
  os_uint32 xtalFreq1,
  os_uint32 xtalFreq2,
  os_uint32 reqFreq,
  os_uint32* actualFreq,
  os_uint32* clockWord)
{
  unsigned long pDiv = 1; /* P divide is 1 because P pin is floating */
  unsigned long T = 0x0; /* Make TEST output low */
  const unsigned long nVal[8] = { 1, 2, 3, 4, 5, 6, 8, 16 };
  unsigned long xtal, bestXtal = 0; /* prevent compiler whining */
  unsigned long mMin, mMax;
  unsigned long m, bestM = 0, n, nIdx, bestNIdx = 0; /* prevent compiler whining */
  unsigned long minFreq, maxFreq;
  unsigned long refFreq;
  unsigned long f, bestFreq = 0; /* prevent compiler whining */
  unsigned long err, bestErr = (unsigned long)-1;
  
  for (xtal = 0; xtal < 2; xtal++) {
    refFreq = xtal ? xtalFreq2 : xtalFreq1;
    mMin = (490000000 * pDiv + refFreq - 1) / refFreq;
    mMax = 640000000 * pDiv / refFreq;
    minFreq = refFreq * mMin / (16 * pDiv) - 1000;
    maxFreq = refFreq * mMax / pDiv + 1000;
    if (reqFreq < minFreq || reqFreq > maxFreq) {
      continue;
    }
    for (m = mMin; m <= mMax; m++) {
      for (nIdx = 0; nIdx < 8; nIdx++) {
	n = nVal[nIdx];
	f = refFreq * m / (n * pDiv);
	if (f > reqFreq) {
	  err = f - reqFreq;
	} else {
	  err = reqFreq - f;
	}
	if (err < bestErr) {
	  bestErr = err;
	  bestFreq = f;
	  bestM = m;
	  bestNIdx = nIdx;
	  bestXtal = xtal;
	}
      }
    }
  }
  
  if (bestM == 0) {
    /* Failed to find a suitable value */
    return MethodResult_InvalidParameter;
  }
  
  *actualFreq = bestFreq;
  *clockWord = (bestXtal << 17) | (T << 15) | (bestNIdx << 12) | (bestNIdx << 9) | (bestM << 0);
  
  return MethodResult_Success;
}

/*
** Bits 0-15 of the returned clockWord are the divider value and
** bits 16-31 are the multiplier value. Max. allowable value of
** 268e6 for 'refFreq' is to prevent arithmetic overflow in calculation
** of 'f'.
*/
static MethodResult
methodClockWordXrc4fxLclk(
  DeviceContext* pDevCtx,
  os_uint32 refFreq,
  os_uint32 reqFreq,
  os_uint32* actualFreq,
  os_uint32* clockWord)
{
  const unsigned long minFreq = 31999000, maxFreq = 210001000;
  unsigned long m, bestM = 0, d, bestD = 0; /* prevent compiler whining */
  unsigned long f, bestFreq = 0; /* prevent compiler whining */
  unsigned long err, bestErr = (unsigned long)-1;
  
  if (refFreq > 268000000) {
    return MethodResult_InvalidParameter;
  }
  if (reqFreq < minFreq || reqFreq > maxFreq) {
    return MethodResult_InvalidParameter;
  }
  
  /* Won't overflow assuming refFreq < 268e6 */
  for (d = 1; d <= 32; d++) {
    for (m = 2; m <= 32; m++) {
      f = (((refFreq >> 1) * m / d) << 1);
      if (reqFreq < minFreq || reqFreq > maxFreq) {
	continue;
      }
      if (f > reqFreq) {
	err = f - reqFreq;
      } else {
	err = reqFreq - f;
      }
      if (err < bestErr) {
	bestErr = err;
	bestFreq = f;
	bestM = m;
	bestD = d;
      }
    }
  }
  
  if (bestErr == (unsigned long)-1) {
    /* Failed to find a suitable value */
    return MethodResult_InvalidParameter;
  }
  
  *actualFreq = bestFreq;
  *clockWord = ((bestD - 1) << 16) | (bestM - 1);
  os_dbgprint(6, ("methodClockWordXrc4fxLclk: bestFreq=%8.8lx clockWord=%8.8lx\n",
		  (unsigned long) bestFreq, (unsigned long) *clockWord));
  
  return MethodResult_Success;
}

MethodResult
methodClockWordXrc(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  os_uint32 fref;

  if (!(pDevCtx->info.modifications & ADMXRC_MOD_CLOCKSWAP)) {
    /*
    ** The VCLK1 and MCLK swap modification is not present
    ** on this board.
    */
    clockIndex = clockIndex ^ 1;
  }
  if (pDevCtx->info.clockType == 0) {
    fref = 16666667;
  } else {
    fref = 14318181;
  }
  return methodClockWordIcd2061(pDevCtx, clockIndex, fref, freq, actual, clockWord);
}

MethodResult
methodClockWordXrc2(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  os_uint32 fref;

  if (pDevCtx->info.clockType == 0) {
    fref = 16666667;
  } else {
    fref = 14318181;
  }
  return methodClockWordIcd2061(pDevCtx, clockIndex, fref, freq, actual, clockWord);
}

MethodResult
methodClockWordXpl(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  return methodClockWordIcs307(pDevCtx, TRUE, 14318181, freq, actual, clockWord);
}

MethodResult
methodClockWordXrc4ls(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  if (clockIndex == 0) {
    return methodClockWordIcs307(pDevCtx, FALSE, 25000000, freq, actual, clockWord);
  } else {
    return methodClockWordIcs8430_61(pDevCtx, 25000000, freq, actual, clockWord);
  }
}

MethodResult
methodClockWordAdpedev(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  if (clockIndex == 0) {
    return methodClockWordIcs307(pDevCtx, FALSE, pDevCtx->clock.refClk.adpedev.lclk, freq, actual, clockWord);
  } else {
    return methodClockWordIcs8430_61(pDevCtx, pDevCtx->clock.refClk.adpedev.mclk, freq, actual, clockWord);
  }
}

MethodResult
methodClockWordXrc4fx(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  if (clockIndex == 0) {
    return methodClockWordXrc4fxLclk(pDevCtx, pDevCtx->clock.refClk.xrc4fx.lclk, freq, actual, clockWord);
  } else {
    return methodClockWordIcs843034_01(pDevCtx,
				       pDevCtx->clock.refClk.xrc4fx.mclkA,
				       pDevCtx->clock.refClk.xrc4fx.mclkB,
				       freq,
				       actual,
				       clockWord);
  }
}

MethodResult
methodClockWordXrc5lx(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  if (clockIndex == 0) {
    return methodClockWordIcs307(pDevCtx,
				 FALSE,
				 pDevCtx->clock.refClk.xrc5lx.lclk,
				 freq,
				 actual,
				 clockWord);
  } else {
    return methodClockWordIcs8430_61(pDevCtx,
				     pDevCtx->clock.refClk.xrc5lx.mclk,
				     freq,
				     actual,
				     clockWord);
  }
}

MethodResult
methodClockWordXrc5lxa(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  if (clockIndex == 0) {
    return methodClockWordXrc4fxLclk(pDevCtx,
				     pDevCtx->clock.refClk.xrc5lxa.lclk,
				     freq,
				     actual,
				     clockWord);
  } else {
    return methodClockWordIcs8430_61(pDevCtx,
				     pDevCtx->clock.refClk.xrc5lxa.mclk,
				     freq,
				     actual,
				     clockWord);
  }
}

/*
** -----------------------------------------------------------------------
** MethodClockprog
** -----------------------------------------------------------------------
*/

MethodResult
methodClockProgPlxbased(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  os_dbgprint(3, ("methodClockProgPlxbased: entered, bBefore = %d\n", (int) bBefore));

  if (bBefore) {
    int i;
    os_uint8 d;

    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
  
    for (i = 0; i < 6; i++) {
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 2);
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 3);
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    }
  
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 2);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 1);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 1);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    
    /* now send the data bits */
    for (i = 0; i < 24; i++) {
      if (clockWord & 0x1U) {
	d = 0;
      } else {
	d = 2;
      }
      clockWord >>= 1;
      
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, (os_uint8) (d | 0x1U));
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, (os_uint8) d);
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, (os_uint8) (d ^ 0x2U));
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, (os_uint8) (d ^ 0x2U) | 0x1U);
      os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    }

    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 2);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.control.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.control.plxbased->CCON);
  } else {
    if (os_read_register_8(&pDevCtx->hw.control.plxbased->CCON) & 0x10) {
      return MethodResult_GeneralFailure;
    }
  }

  return MethodResult_Success;
}

MethodResult
methodClockProgXplrev0(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  XPLSPACE1REGS* regs;
  os_uint32      val32;

  os_dbgprint(3, ("methodClockProgXplrev0: entered bBefore = %d\n", (int) bBefore));

  regs = pDevCtx->hw.control.xpl;

  if (bBefore) {
    os_write_register_le32(&regs->CLOCK, (clockWord << 8) | 0x1U);
  } else {
    val32 = os_read_register_le32(&regs->CLOCK);
    if (!(val32 & 0x1)) {
      return MethodResult_Timeout;
    }
  }

  return MethodResult_Success;
}

MethodResult
methodClockProgXplrev2(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  XPLSPACE1REGS* regs;
  os_uint32      newclkdiv, clockreg;
  os_uint32      val32;

  os_dbgprint(3, ("methodClockProgXplrev2: entered\n"));

  regs = pDevCtx->hw.control.xpl;

  if (freq < 50001000) {
    /* LCLK divide of 1 for 0-50MHz */
    newclkdiv = 2;
  } else if (freq < 75001000) {
    /* LCLK divide of 1.5 for 50-75MHz */
    newclkdiv = 1;
  } else {
    /* LCLK divide of 2 for 75+MHz */
    newclkdiv = 0;
  }

  if (bBefore) {
    os_uint32 oldclkdiv;

    oldclkdiv = (os_read_register_le32(&regs->CLOCK) >> 2) & 0x3U;
    if (oldclkdiv > newclkdiv) {
      /* Increasing the frequency; change division factor at the same time as changing frequency */
      clockreg = (clockWord << 8) | (newclkdiv << 2);
    } else {
      /* Decreasing the frequency; change division factor AFTER changing frequency */
      clockreg = (clockWord << 8) | (oldclkdiv << 2);
    }
  
    os_write_register_le32(&regs->CLOCK, clockreg | 0x1U);
    os_read_register_le32(&regs->CLOCK);
  } else {
    val32 = os_read_register_le32(&regs->CLOCK);
    if (!(val32 & 0x1)) {
      return MethodResult_Timeout;
    }
  
    /* Decreasing the frequency; change division factor AFTER changing frequency */
    clockreg = (clockWord << 8) | (newclkdiv << 2);
    os_write_register_le32(&regs->CLOCK, clockreg);
    os_read_register_le32(&regs->CLOCK);
  }

  return MethodResult_Success;
}

MethodResult
methodClockProgXrc4ls(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  CONTROLXRC4LS* regs;
  os_uint32 val32;

  os_dbgprint(3, ("methodClockProgXrc4ls: entered, freq=%ld bBefore=%d\n",
		  (unsigned long) freq, (int) bBefore));
	
  regs = pDevCtx->hw.control.xrc4ls;

  if (bBefore) {
    if (clockIndex == 0) {
      os_write_register_le32(&regs->LCLOCK, clockWord | 0x80000000U);
      os_read_register_le32(&regs->LCLOCK);
    } else {
      os_write_register_le32(&regs->MCLOCK, clockWord | 0xC0000000U);
      os_read_register_le32(&regs->MCLOCK);
    }
  } else {
    if (clockIndex == 0) {
      val32 = os_read_register_le32(&regs->LCLOCK);
    } else {
      val32 = os_read_register_le32(&regs->MCLOCK);
    }
    if (!(val32 & 0x80000000U)) {
      return MethodResult_Timeout;
    }
  }
  return MethodResult_Success;
}

MethodResult
methodClockProgXrc4fx(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  XRC4FXSPACE1REGS* regs;
  os_uint32 val32, mask;
  
  os_dbgprint(3, ("methodClockProgXrc4fx: entered, freq=%ld bBefore=%d\n",
		  (unsigned long) freq, (int) bBefore));
  
  regs = pDevCtx->hw.control.xrc4fx;
  if (bBefore) {
    if (clockIndex == 0) {
      os_write_register_le32(&regs->LCLKCTL, clockWord | (0x1U << 31));
      os_read_register_le32(&regs->LCLKSTAT3);
    } else {
      os_write_register_le32(&regs->USERMGT,
			   (0x1U << 31) | ((clockWord & 0x20000U) << 13) | (clockWord & 0x1ffffU));
      os_read_register_le32(&regs->USERMGT);
    }
  } else {
    if (clockIndex == 0) {
      val32 = os_read_register_le32(&regs->LCLKSTAT3);
      mask = (val32 & 0x4U) ? 0x2U : 0x1U;
      if (!(val32 & mask)) {
	/* DCM not locked */
	return MethodResult_GeneralFailure;
      }
      if (val32 & 0x28U) {
	/* Control logic busy or write of DCM parameters timed out */
	return MethodResult_Timeout;
      }
      if (val32 & 0x10U) {
	/* Error */
	return MethodResult_GeneralFailure;
      }
    } else {
      val32 = os_read_register_le32(&regs->USERMGT);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    }
  }
  return MethodResult_Success;
}

MethodResult
methodClockProgAdpedev(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  ADPEDEVSPACE1REGS* regs;
  os_uint32 val32;
  
  os_dbgprint(3, ("methodClockProgAdpedev: entered, freq=%ld bBefore=%d\n",
		  (unsigned long) freq, (int) bBefore));
  
  regs = pDevCtx->hw.control.adpedev;
  if (bBefore) {
    if (clockIndex == 0) {
      os_write_register_le32(&regs->LCLK, (0x1U << 31) | (clockWord & 0xffffffU));
      os_read_register_le32(&regs->LCLK);
    } else {
      os_write_register_le32(&regs->MCLK, (0x3U << 30) | (clockWord & 0x3fffU));
      os_read_register_le32(&regs->LCLK);
    }
  } else {
    if (clockIndex == 0) {
      val32 = os_read_register_le32(&regs->LCLK);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    } else {
      val32 = os_read_register_le32(&regs->MCLK);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    }
  }
  return MethodResult_Success;
}

MethodResult
methodClockProgXrc5lx(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  XRC5LXSPACE1REGS* regs;
  os_uint32 val32;
  
  os_dbgprint(3, ("methodClockProgXrc5lx: entered, freq=%ld bBefore=%d\n",
		  (unsigned long) freq, (int) bBefore));
  
  regs = pDevCtx->hw.control.xrc5lx;
  if (bBefore) {
    if (clockIndex == 0) {
      os_write_register_le32(&regs->LCLKCTL, (0x1U << 31) | (clockWord & 0xffffffU));
      os_read_register_le32(&regs->LCLKCTL);
    } else {
      os_write_register_le32(&regs->MCLKCTL, (0x3U << 30) | (clockWord & 0x3fffU));
      os_read_register_le32(&regs->MCLKCTL);
    }
  } else {
    if (clockIndex == 0) {
      val32 = os_read_register_le32(&regs->LCLKCTL);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    } else {
      val32 = os_read_register_le32(&regs->MCLKCTL);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    }
  }
  return MethodResult_Success;
}

MethodResult
methodClockProgXrc5lxa(
  DeviceContext* pDevCtx,
  unsigned int clockIndex,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  XRC5LXASPACE1REGS* regs;
  os_uint32 val32, mask;
  
  os_dbgprint(3, ("methodClockProgXrc5lxa: entered, freq=%ld bBefore=%d\n",
		  (unsigned long) freq, (int) bBefore));
  
  regs = pDevCtx->hw.control.xrc5lxa;
  if (bBefore) {
    if (clockIndex == 0) {
      os_write_register_le32(&regs->LCLKCTL, clockWord | (0x1U << 31));
      os_read_register_le32(&regs->LCLKSTAT3);
    } else {
      if (pDevCtx->devId.pci.revision >= 0x1U) {
	os_write_register_le32(&regs->MCLKCTL, (0x3U << 30) | (clockWord & 0x3fffU));
      } else {
	/* Bug in ADM-XRC-5LXA Bridge with PCI rev 00 - clockword field turns out to be in bits [20:7] */
	os_write_register_le32(&regs->MCLKCTL, (0x3U << 30) | ((clockWord & 0x3fffU) << 7U));
      }
      os_read_register_le32(&regs->MCLKCTL);
    }
  } else {
    if (clockIndex == 0) {
      val32 = os_read_register_le32(&regs->LCLKSTAT3);
      mask = (val32 & 0x4U) ? 0x2U : 0x1U;
      if (!(val32 & mask)) {
	/* DCM not locked */
	return MethodResult_GeneralFailure;
      }
      if (val32 & 0x28U) {
	/* Control logic busy or write of DCM parameters timed out */
	return MethodResult_Timeout;
      }
      if (val32 & 0x10U) {
	/* Error */
	return MethodResult_GeneralFailure;
      }
    } else {
      val32 = os_read_register_le32(&regs->MCLKCTL);
      if (!(val32 & 0x80000000U)) {
	return MethodResult_Timeout;
      }
    }
  }
  return MethodResult_Success;
}

